import numpy as np
import math
import torch
import tensorflow as tf 

class CarCost(object):

    def __init__(self, traj_size=30, pre_process=None, use_delta=False):
        self.traj_size = traj_size
        self.multiply_factor = 1
        self.use_delta = use_delta
        if pre_process is not None:
            pre_process = pre_process[:1000]
            detail = self.cost_function_pt(torch.Tensor(pre_process))
            detail = detail.data.cpu().numpy()
            weight = np.array([np.sqrt(np.var(detail[..., i])) for i in range(detail.shape[-1])])
            # weight = np.array([np.max(detail[..., i]) for i in range(detail.shape[-1])])
            weight = 1 / weight
            self.multiply_factor = weight[np.newaxis, np.newaxis, :]
            print("Cost Function Preprocessed with : ", self.multiply_factor)

    @staticmethod
    def lane_function(x, c):
        if type(c)==torch.Tensor:
            res = torch.zeros_like(c[..., 0])
        else: 
            res = tf.zeros_like(c[..., 0])
        for j in range(c.shape[-1]):
            res += c[..., j] * (x ** j)
        return res

    @staticmethod
    def signed_distance(x1, x2, r1=1, r2=1):
        return np.sqrt(np.power((x1[..., 0] - x2[..., 0]) / r1, 4) + np.power((x1[..., 1] - x2[..., 1]) / r2, 4))

    @staticmethod
    def signed_distance_pt(x1, x2, r1=1, r2=1):
        return torch.sqrt(torch.pow((x1[..., 0] - x2[..., 0]) / r1, 4) + torch.pow((x1[..., 1] - x2[..., 1]) / r2, 4) + 1e-8)

    @staticmethod
    def lane_direction(x, c):
        if c.shape[-1] == 5:
            return np.arctan((c[..., 4] * 4 * x ** 3 + c[..., 3] * 3 * x ** 2 + c[..., 2] * 2 * x + c[..., 1]))
        else:
            return np.arctan((c[..., 3] * 3 * x ** 2 + c[..., 2] * 2 * x + c[..., 1]))

    @staticmethod
    def lane_direction_pt(x, c):
        if c.shape[-1] == 5:
            return torch.atan(1.0/(c[..., 4] * 4 * x ** 3 + c[..., 3] * 3 * x ** 2 + c[..., 2] * 2 * x + c[..., 1] + 1e-8))
        else:
            return torch.atan(1.0/(c[..., 3] * 3 * x ** 2 + c[..., 2] * 2 * x + c[..., 1] + 1e-8))

    @staticmethod
    def sabs(x, p):
        return np.sqrt(np.power(x, 2) + np.power(p, 2)) - p

    @staticmethod
    def sabs_pt(x, p):
        return torch.sqrt(torch.pow(x, 2) + p * p) - p

    def lane_distance(self, x, lane):
        m = self.lane_function(x[..., 0], lane)
        return m - x[..., 1]
        # return np.minimum(np.minimum(m - x[..., 1], m+4 - x[..., 1]), m-4 - x[..., 1])**2

class CarCostSingleIdentical(CarCost):

    def cost_function_pt(self, x, y):
        return torch.cat([x, y], dim=-1)

class CarCostSingleDefaultISEE(CarCost): # Repara enhanced

    # Input matrix : [batch_size * num_frame * 37]
    # cost_function output: [batch_size * num_frame * 10]

    def cost_obstacle(self, x, y, egoRadius=1):
        c = torch.tensor(0.0, device=x.device)
        idx = 0
        for i in range(0, 5):
            c = c + torch.exp(-self.signed_distance_pt(x[..., 0:2],
                    y[..., idx:(idx + 2)], egoRadius, egoRadius))
            idx += 4
        return c

    def cost_function_pt(self, data_status, data_extend):

        x = data_status[..., 0:6]
        u = data_status[..., 6:8]
        y = data_extend[..., :20]
        goal = data_extend[..., 20:22]
        lane = data_extend[..., -5:-1]
        speedLimit = data_extend[..., -1]
        delta_padding = torch.zeros((x.shape[0], 1), device=data_status.device)

        output_list = torch.stack((self.sabs_pt(x[..., 1] - goal[..., 1], 0.1),  # Cost_goal lat - y
                                self.sabs_pt(x[..., 0] - goal[..., 0], 0.1),  # Cost_goal long - x
                                self.sabs_pt(self.lane_distance(x, lane), 0.1),  # Cost_lane
                                self.sabs_pt(x[..., 3] - speedLimit, 2),  # Cost_speed
                                torch.cat((torch.pow(x[..., :-1, 4] if self.use_delta else 
                                u[..., :-1, 0] - x[..., :-1, 4], 2), delta_padding), 1),  # Cost_delta_u long
                                torch.cat((torch.pow(x[..., :-1, 5] if self.use_delta else 
                                u[..., :-1, 1] - x[..., :-1, 5], 2), delta_padding), 1),  # Cost_delta_u lat
                                self.sabs_pt(x[..., 2] + self.lane_direction_pt(x[..., 0], lane) + np.pi / 2,
                                          0.02),  # Cost_direction
                                self.cost_obstacle(x, y),  # Cost_obstacle
                                torch.pow(u[..., 0], 2),  # Cost_u long
                                torch.pow(u[..., 1], 2)), 2)  # Cost_u lat
        return output_list

class CarDynamic(object):

    def __init__(self, traj_size=30, multiple_factor=[1,1]):
        self.traj_size = traj_size
        self.multiple_factor = multiple_factor
        self.state_size = 6
        self.action_size = 2
        self.sa_size = 8

class CarDynamicDefault(CarDynamic):

    # Bicycle model (https://hal-polytechnique.archives-ouvertes.fr/hal-01520869/document)
    # For single-agent :
    # x0 : batch_size * 6; us : batch_size * num_frame * 2

    def car_dynamic_naive(self, x, us):

        d = 3.
        Kus = 0.043
        dt = 0.1
        xs = tf.TensorArray(tf.float32, size=self.traj_size)
        xs = xs.write(0, x)
        for i in range(self.traj_size - 1):
            u = us[:, i, :]
            front_wheel_rolling_dist = dt * x[..., 3]
            back_wheel_rolling_dist = d + front_wheel_rolling_dist * tf.cos(u[..., 0] / self.multiple_factor[0]) \
                                      - tf.sqrt(d * d - np.power(front_wheel_rolling_dist * tf.sin(u[..., 0] / self.multiple_factor[0]), 2))
            x_n = tf.stack(
                [x[..., 0] + back_wheel_rolling_dist * tf.sin(x[..., 2]),
                 x[..., 1] + back_wheel_rolling_dist * tf.cos(x[..., 2]),
                 x[..., 2] + tf.asin(
                     tf.sin(u[..., 0] / self.multiple_factor[0]) * front_wheel_rolling_dist / (d + Kus * np.power(x[..., 3], 2) / 9.8)),
                 x[..., 3] + u[..., 1] / self.multiple_factor[1] * dt, u[..., 0], u[..., 1]], axis=1)
            xs = xs.write(i + 1, x_n)
            x = x_n
        return tf.transpose(xs.stack(), [1, 0, 2])

    def car_dynamic(self, s, u, ori_angle=-100):

        if len(u.shape) == 2:
            s = np.expand_dims(s, axis=0)
            u = np.expand_dims(u, axis=0)
            expand = True
        else:
            expand = False
        if ori_angle > -10:
            s[:, 2] = ori_angle + math.pi / 2
        x = np.empty((u.shape[0], u.shape[1], 6))
        x[:, 0, :] = s
        for i in range(u.shape[1] - 1):
            dt = 0.1
            d = 3.
            Kus = 0.043
            s = x[:, i, :]
            front_wheel_rolling_dist = dt * s[:, 3]
            back_wheel_rolling_dist \
                = d + front_wheel_rolling_dist * np.cos(u[:, i, 0] / self.multiple_factor[0]) \
                  - np.sqrt(d * d - np.power(front_wheel_rolling_dist *
                                             np.sin(u[:, i, 0] / self.multiple_factor[0]), 2))

            x[:, i + 1, :] = np.array(
                [s[:, 0] + back_wheel_rolling_dist * np.sin(s[:, 2]),
                 s[:, 1] + back_wheel_rolling_dist * np.cos(s[:, 2]),
                 s[:, 2] + np.arcsin(np.sin(u[:, i, 0] / self.multiple_factor[0]) * front_wheel_rolling_dist /
                                     (d + Kus * np.power(s[:, 3], 2) / 9.8)),
                 s[:, 3] + u[:, i, 1] / self.multiple_factor[1] * dt,
                 u[:, i, 0],
                 u[:, i, 1]]
            ).T
        if expand:
            x = x.squeeze(axis=0)
        return x

    def car_dynamic_step(self, s, u):

        dt = 0.1
        d = 3.
        Kus = 0.043
        if type(s) == torch.Tensor:

            front_wheel_rolling_dist = dt * s[..., 3]
            back_wheel_rolling_dist \
                = d + front_wheel_rolling_dist * torch.cos(u[..., 0] / self.multiple_factor[0]) \
                - torch.sqrt(d * d - torch.pow(front_wheel_rolling_dist * torch.sin(u[..., 0] / self.multiple_factor[0]), 2))
            return torch.stack(
                [s[..., 0] + back_wheel_rolling_dist * torch.sin(s[..., 2]),
                s[..., 1] + back_wheel_rolling_dist * torch.cos(s[..., 2]),
                s[..., 2] + torch.asin(torch.sin(u[..., 0] / self.multiple_factor[0]) * front_wheel_rolling_dist /
                                    (d + Kus * torch.pow(s[..., 3], 2) / 9.8)),
                s[..., 3] + u[..., 1] / self.multiple_factor[1] * dt,
                u[..., 0],
                u[..., 1]], dim=1)
        else:
            front_wheel_rolling_dist = dt * s[..., 3]
            back_wheel_rolling_dist \
                = d + front_wheel_rolling_dist * np.cos(u[..., 0] / self.multiple_factor[0]) \
                - np.sqrt(d * d - np.power(front_wheel_rolling_dist *
                                            np.sin(u[..., 0] / self.multiple_factor[0]), 2))
            x = np.array(
                [s[..., 0] + back_wheel_rolling_dist * np.sin(s[..., 2]),
                s[..., 1] + back_wheel_rolling_dist * np.cos(s[..., 2]),
                s[..., 2] + np.arcsin(np.sin(u[..., 0] / self.multiple_factor[0]) * front_wheel_rolling_dist /
                                    (d + Kus * np.power(s[..., 3], 2) / 9.8)),
                s[..., 3] + u[..., 1] / self.multiple_factor[1] * dt,
                u[..., 0],
                u[..., 1]]
            ).T
            return x

    def car_dynamic_pt(self, s, u):

        x = torch.cat((s[:, 0, :4].unsqueeze(1), torch.zeros((s.shape[0], 1, 2), device=s.device)), 2)
        for i in range(u.shape[1] - 1):
            dt, d, kus = 0.1, 3, 0.043
            s = x[:, -1]
            front_wheel_rolling_dist = dt * s[:, 3]
            back_wheel_rolling_dist = d + front_wheel_rolling_dist * torch.cos(u[..., i, 0] / self.multiple_factor[0]) \
                                      - torch.sqrt(d * d - torch.pow(front_wheel_rolling_dist * torch.sin(u[..., i, 0] / self.multiple_factor[0]), 2))

            x_new = torch.stack(
                (s[:, 0] + back_wheel_rolling_dist * torch.sin(s[:, 2]),
                 s[:, 1] + back_wheel_rolling_dist * torch.cos(s[:, 2]),
                 s[:, 2] + torch.asin(
                     torch.sin(u[..., i, 0] / self.multiple_factor[0]) * front_wheel_rolling_dist / (d + kus * torch.pow(s[:, 3], 2) / 9.8)),
                 s[:, 3] + u[..., i, 1] / self.multiple_factor[1] * dt,
                 u[..., i, 0],
                 u[..., i, 1]), 1).unsqueeze(1)
            x = torch.cat((x, x_new), 1)
        return torch.cat((x, u), 2)

class CarDynamicReparam(CarDynamic):

    # Reparameterized dynamic function by using delta action as action
    # For single-agent :
    # x0 : batch_size * 6; us : batch_size * num_frame * 2

    def car_dynamic_pt(self, s, u):

        x = (s[:, 0, :6] if len(s.shape)==3 else s[:, :6]).unsqueeze(1)
        dt = 0.1
        for i in range(u.shape[1] - 1):
            dt, d, kus = 0.1, 3, 0.043
            s = x[:, -1]
            front_wheel_rolling_dist = dt * s[:, 3]
            back_wheel_rolling_dist = d + front_wheel_rolling_dist * torch.cos((s[..., 4] + u[..., i, 0]) / self.multiple_factor[0]) \
                                      - torch.sqrt(d * d - torch.pow(front_wheel_rolling_dist * torch.sin((s[..., 4] + u[..., i, 0]) / self.multiple_factor[0]), 2))

            x_new = torch.stack(
                (s[:, 0] + back_wheel_rolling_dist * torch.sin(s[:, 2]),
                 s[:, 1] + back_wheel_rolling_dist * torch.cos(s[:, 2]),
                 s[:, 2] + torch.asin(
                     torch.sin((s[..., 4] + u[..., i, 0]) / self.multiple_factor[0]) * front_wheel_rolling_dist / (d + kus * torch.pow(s[:, 3], 2) / 9.8)),
                 s[:, 3] + (s[..., 5] + u[..., i, 1]) / self.multiple_factor[1] * dt,
                 (s[..., 4] + u[..., i, 0]),
                 (s[..., 5] + u[..., i, 1])), 1).unsqueeze(1)
            x = torch.cat((x, x_new), 1)
        return torch.cat((x, u), 2)


    def car_dynamic(self, s, u, ori_angle=-100):

        if len(u.shape) == 2:
            s = np.expand_dims(s, axis=0)
            u = np.expand_dims(u, axis=0)
            expand = True
        else:
            expand = False
        if ori_angle > -10:
            s[:, 2] = ori_angle + math.pi / 2
        x = np.empty((u.shape[0], u.shape[1], 6))
        x[:, 0, :] = s
        dt = 0.1
        for i in range(u.shape[1] - 1):
            dt = 0.1
            d = 3.
            Kus = 0.043
            s = x[:, i, :]
            front_wheel_rolling_dist = dt * s[:, 3]
            back_wheel_rolling_dist \
                = d + front_wheel_rolling_dist * np.cos((s[:, 4] + u[:, i, 0]) / self.multiple_factor[0]) \
                  - np.sqrt(d * d - np.power(front_wheel_rolling_dist *
                                             np.sin((s[:, 4] + u[:, i, 0]) / self.multiple_factor[0]), 2))

            x[:, i + 1, :] = np.array(
                [s[:, 0] + back_wheel_rolling_dist * np.sin(s[:, 2]),
                 s[:, 1] + back_wheel_rolling_dist * np.cos(s[:, 2]),
                 s[:, 2] + np.arcsin(np.sin((s[:, 4] + u[:, i, 0]) / self.multiple_factor[0]) * front_wheel_rolling_dist /
                                     (d + Kus * np.power(s[:, 3], 2) / 9.8)),
                 s[:, 3] + (s[:, 5] + u[:, i, 1]) / self.multiple_factor[1] * dt,
                 (s[:, 4] + u[:, i, 0]),
                 s[:, 5] + u[:, i, 1]]
            ).T
        if expand:
            x = x.squeeze(axis=0)
        return x

    def car_dynamic_step(self, s, u):

        dt = 0.1
        d = 3.
        Kus = 0.043
        if type(s) == torch.Tensor:

            front_wheel_rolling_dist = dt * s[..., 3]
            back_wheel_rolling_dist \
                = d + front_wheel_rolling_dist * torch.cos((s[:, 4] + u[:, 0]) / self.multiple_factor[0]) \
                - torch.sqrt(d * d - torch.pow(front_wheel_rolling_dist * torch.sin((s[:, 4] + u[:, 0]) / self.multiple_factor[0]), 2))
            return torch.stack(
                [s[..., 0] + back_wheel_rolling_dist * torch.sin(s[..., 2]),
                s[..., 1] + back_wheel_rolling_dist * torch.cos(s[..., 2]),
                s[..., 2] + torch.asin(torch.sin((s[:, 4] + u[:, 0]) / self.multiple_factor[0]) * front_wheel_rolling_dist /
                                    (d + Kus * torch.pow(s[..., 3], 2) / 9.8)),
                s[..., 3] + (s[:, 5] + u[:, 1]) / self.multiple_factor[1] * dt,
                s[:, 4] + u[:, 0],
                s[:, 5] + u[:, 1]], dim=1)
        else:
            front_wheel_rolling_dist = dt * s[..., 3]
            back_wheel_rolling_dist \
                = d + front_wheel_rolling_dist * np.cos((s[:, 4] + u[:, 0]) / self.multiple_factor[0]) \
                - np.sqrt(d * d - np.power(front_wheel_rolling_dist *
                                            np.sin((s[:, 4] + u[:, 0]) / self.multiple_factor[0]), 2))
            x = np.array(
                [s[..., 0] + back_wheel_rolling_dist * np.sin(s[..., 2]),
                s[..., 1] + back_wheel_rolling_dist * np.cos(s[..., 2]),
                s[..., 2] + np.arcsin(np.sin((s[:, 4] + u[:, 0]) / self.multiple_factor[0]) * front_wheel_rolling_dist /
                                    (d + Kus * np.power(s[..., 3], 2) / 9.8)),
                s[..., 3] + (s[:, 5] + u[:, 1]) / self.multiple_factor[1] * dt,
                s[:, 4] + u[:, 0],
                s[:, 5] + u[:, 1]]
            ).T
            return x
