# Add the current path to python
import sys
sys.path.append('.')
import traceback
# Required if you need to output gif
import matplotlib
matplotlib.use('Agg')
from matplotlib import rcParams
rcParams.update({'figure.autolayout': True})
font_size = 18
matplotlib.rc('xtick', labelsize=font_size) 
matplotlib.rc('ytick', labelsize=font_size) 

import argparse
import numpy as np
import torch
import os
import torch.utils.data as tdata
import cost_dynamic as cost_dynamic
import matplotlib.pyplot as plt
from progressbar import ETA, Bar, Percentage, ProgressBar
import util_torch
import network as NNModel
from torch.utils.tensorboard import SummaryWriter, FileWriter
import datetime
try:
    import tianshou as ts 
except:
    raise ImportError

class LangevinMEIOC(torch.nn.Module):

    """
    Langevin sampler for energy based model with dynamic function supported.
    """
    def __init__(self, net, dynamic, config, tb_writer):

        super().__init__()
        self.des_model, self.dynamic = net, dynamic 
        self.C, self.tb_writer = config, tb_writer

    def forward(self, x_init, status, extend=None, test_epoch=-1, short_run=-1):

        u = torch.autograd.Variable(x_init[..., self.C.his_size:, -self.C.action_size:], requires_grad=True)
        x_his = status[..., :self.C.his_size, :]
        x_future = status[..., self.C.his_size:, :]
        n_step = short_run if short_run > 0 else self.C.sample_steps
        x_full = []
        if self.C.print_level > 5 and test_epoch!=-1:
            fig = plt.figure()
            ax = fig.add_subplot()
            c_list, grad_list, du_list = np.zeros(n_step), np.zeros(n_step), np.zeros(n_step)
            x_full, error = [], []
        for i in range(n_step):
            x_new = torch.cat((x_his, self.dynamic(x_future, u)), 1)
            cost = self.des_model(x_new, extend)
            grad = torch.autograd.grad(cost.sum(), [u], retain_graph=True)[0]
            noise = torch.randn(u.shape, device=status.device) if i < self.C.noise_steps \
                else torch.zeros_like(u, device=status.device)
            du = torch.clamp((0.5 * self.C.delta * self.C.delta * (u / self.C.sigma / self.C.sigma + grad)),
                            -self.C.default_limit, self.C.default_limit) + self.C.delta * noise
            u = u - du
            if self.C.print_level > 5 and test_epoch!=-1:
                pnt_u = u.cpu().data.numpy()
                ax.plot(pnt_u[-1, :, 0], c=(1, 1-i/n_step, 1-i/n_step))
                ax.plot(pnt_u[-1, :, 1], c=(1-i/n_step, 1-i/n_step, 1))
                c_list[i] = torch.mean(cost).cpu().data.numpy()
                grad_list[i] = torch.mean(torch.abs(grad)).cpu().data.numpy()
                du_list[i] = torch.mean(torch.abs(du)).cpu().data.numpy()
                error.append(torch.mean(util_torch.l2_error(status, x_new, True)[:, -1]).cpu().data.numpy())
            if self.C.print_level > 99:
                x_full.append(x_new.detach())

        if self.C.print_level > 5 and test_epoch!=-1:
            pnt_d = status[-1, self.C.his_size:, -2:].cpu().data.numpy()
            ax.plot(pnt_d[:,0], c=(1,0,0), linestyle='dashed')
            ax.plot(pnt_d[:,1], c=(0,0,1), linestyle='dashed')
            self.tb_writer.add_figure("test/langevin", fig, test_epoch)
            self.tb_writer.add_figure("test/cost", util_torch.EvaluationMetric._print_graph(c_list), test_epoch)
            self.tb_writer.add_figure("test/gradient", util_torch.EvaluationMetric._print_graph(grad_list), test_epoch)
            self.tb_writer.add_figure("test/delta_u", util_torch.EvaluationMetric._print_graph(du_list), test_epoch)
            self.tb_writer.add_figure("test/error_step", util_torch.EvaluationMetric._print_graph(np.sqrt(np.array(error))), test_epoch)

        x_new = torch.cat((x_his, self.dynamic(x_future, u)), 1)
        if self.C.print_level > 99:
            x_full.append(x_new)
            return x_full

        # Do loss calculation 
        # if status.shape[1] == extend.shape[1]: 
        #     obs_cost = self.des_model(status, extend)
        #     syn_cost = self.des_model(x_new, extend)
        #     loss = torch.mean(obs_cost) - torch.mean(syn_cost)
        #     return x_new, loss, [obs_cost, syn_cost]
        # else: 
        return x_new

class IOC_Cooperative(object):

    def __init__(self, config, test=False):

        self.C = config
        
        # Define general requirement for Autonomous Driving.
        self.car_dynamic = getattr(cost_dynamic, "CarDynamic" + ("" if self.C.single else "Multi") + config.dynamic)(self.C.traj_size, config.multiple_factor)
        self.cost_func = getattr(cost_dynamic, "CarCostSingle" + self.C.costFunc)(self.C.traj_size, use_delta=(config.dynamic=='Reparam')).cost_function_pt
        self.des_warpper = lambda s, e: torch.cat([s[..., :6], e])

        # Logistic 
        self.tb_writer = SummaryWriter(log_dir="output/runs/%s" % self.C.category)
        params = [attr for attr in dir(self.C) if not attr.startswith("__") and (type(getattr(self.C, attr))==str 
                    or type(getattr(self.C, attr))==float or type(getattr(self.C, attr))==int)]
        self.tb_writer.add_text("config", str({k : getattr(self.C, k) for k in params}))
        self.device = torch.device(config.device) if hasattr(config, 'device') else \
                      torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

        # Define whole network
        if config.continue_train == True: 
            self.gen_model, self.des_model = torch.load("output/model/%s.pt" % config.category)
            print("Loaded model from %s" % config.category)
        else: 
            self.gen_model, self.des_model = self.create_network() 
        self.langevin = LangevinMEIOC(self.des_model, self.car_dynamic.car_dynamic_pt, config, self.tb_writer)
            
    def create_network(self):
        """
        Create both network
        """
        if self.C.des_net == "None":
            des_net = None
        else: 
            des_net = getattr(NNModel, "Model" + self.C.des_net)(self.C, self.cost_func)
            if self.device.type == "cuda":
                des_net = des_net.cuda()
        if self.C.gen_net == "None":
            gen_net = None
        else: 
            gen_net = getattr(NNModel, "Generator" + self.C.gen_net)(self.C, self.car_dynamic.car_dynamic_step)
            if self.device.type == "cuda":
                gen_net = gen_net.cuda()
        return gen_net, des_net 

    def normalize_weight(self, data):

        cost = torch.zeros(self.C.feature_size, device=self.device)
        data_loader = tdata.DataLoader(data, batch_size=self.C.batch_size)
        for d in data_loader:
            cost += torch.sum(self.cost_func(d['status'], d['extend']), (0, 1))
        cost = 1 / (cost / cost[0])
        # cost[4:6] = 0 
        # cost[7] = 5
        return cost

    def cooperative_process_multi(self, status, extend, test_epoch=-1):

        n, bn = self.C.num_sample, status.shape[0]
        if n == 1: 
            return self.cooperative_process(status, extend, test_epoch)
        else: 
            return self.cooperative_process(status.repeat(n, 1, 1), extend.repeat(n, 1, 1), test_epoch)

    def cooperative_process(self, status, extend, test_epoch=-1):

        # 2.1 Genertor process
        if self.gen_model is None: 
            if self.C.single:
                t_init = torch.randn((status.shape[0], self.C.traj_size, 8), device=self.device) * self.C.sigma
            else:
                t_init = torch.randn((status.shape[0], self.C.agent_size, self.C.future_size, 2), device=self.device) * self.C.sigma
        else: 
            t_init = torch.cat((status[..., :self.C.his_size, :], self.gen_model.sample(status[..., self.C.his_size:, :], 
                            extend[..., self.C.his_size:, :])), 1)

        # 2.2 Langevin process

        if self.des_model is None or self.C.des_net[:4] == "GAIL":
            return t_init, t_init 
        else:
            return t_init, self.langevin(t_init, status, extend, test_epoch)

    def train(self, train_data, test_data):

        if self.des_model is not None:
            if self.C.weight_decay == 0:
                des_optimizer = torch.optim.Adam(self.des_model.parameters(), lr=self.C.lr_des)
            else:
                des_optimizer = torch.optim.AdamW(self.des_model.parameters(), lr=self.C.lr_des, weight_decay=self.C.weight_decay)

            des_scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer=des_optimizer, gamma=self.C.gamma_decay)
            # 1.2 Calculate normalize weight.
            if self.C.costFunc != "Identical":
                self.des_model.weight = self.normalize_weight(train_data)
                print(self.des_model.weight)
        if self.gen_model is not None and not isinstance(self.gen_model, ts.policy.BasePolicy):
            if self.C.weight_decay == 0:
                gen_optimizer = torch.optim.Adam(self.gen_model.parameters(), lr=self.C.lr_gen)
            else:
                gen_optimizer = torch.optim.AdamW(self.gen_model.parameters(), lr=self.C.lr_gen, weight_decay=self.C.weight_decay)
            gen_scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer=gen_optimizer, gamma=self.C.gamma_decay)

        util_torch.basic_check(test_data[:]['status'], self.car_dynamic.car_dynamic_pt, self.C)

        train_loader = tdata.DataLoader(train_data, batch_size=self.C.batch_size, shuffle=True)
        validate_loader = tdata.DataLoader(train_data, batch_size=self.C.batch_size*5, shuffle=False)
        test_loader = tdata.DataLoader(test_data, batch_size=self.C.batch_size*5, shuffle=False)
        
        epoch = self.C.current_epoch if self.C.continue_train else 0
        self.pbar = ProgressBar(maxval=self.C.eval_interval, widgets=["Epoch #%d|" % epoch, Percentage(), Bar(), ETA()])
        self.pbar.start()
        while True:
            for i_batch, b_data in enumerate(train_loader):

                status = b_data['status'].type(torch.float32)
                extend = b_data['extend'].type(torch.float32)

                gen_syn, des_syn = self.cooperative_process_multi(status, extend)


                if self.des_model is not None:

                    if self.C.des_net == "GAIL":

                        labels = torch.cat([torch.zeros((status.shape[0], status.shape[1]), device=status.device),
                                            torch.ones((gen_syn.shape[0], gen_syn.shape[1]), device=status.device)])
                        des_loss, d_out = self.des_model(torch.cat([status, gen_syn]), extend.repeat(self.C.num_sample + 1, 1, 1), labels)
                        d_out = torch.clamp(d_out, 1e-20)
                        obs_cost, syn_cost = -torch.log(d_out[:status.shape[0]]), -torch.log(d_out[status.shape[0]:]) 
                        self.tb_writer.add_scalar("general/d_acc", torch.mean(((d_out > 0.5) == labels.bool()).float()), epoch)

                    else: 
                        obs_cost = self.des_model(status, extend)
                        syn_cost = self.des_model(des_syn, extend.repeat(self.C.num_sample, 1, 1))
                        des_loss = torch.mean(obs_cost) - torch.mean(syn_cost) 
                        learned_theta = list(self.des_model.parameters())[0].cpu().data.numpy() if self.C.des_net == 'Linear' else None
                    des_optimizer.zero_grad()
                    des_loss.backward()
                    des_optimizer.step()
                    des_scheduler.step()
                    self.tb_writer.add_scalar("general/lr_des", des_scheduler.get_last_lr()[0], epoch)
                else:
                    des_loss, obs_cost, syn_cost = torch.tensor(-1).float(), torch.tensor(-1).float(), torch.tensor(-1).float() 

                if self.gen_model is not None:
                    if isinstance(self.gen_model, ts.policy.BasePolicy): 

                        res = self.gen_model.learn(des_syn.detach(), extend.repeat(self.C.num_sample, 1, 1), syn_cost.detach())
                        for k in res: 
                            if type(res[k]) is list:
                                res[k] = np.mean(res[k])
                            self.tb_writer.add_scalar("general/" + k, float(res[k]), epoch)
                    else: 
                        if self.des_model is None:
                            gen_loss = torch.mean(torch.pow(status.repeat(self.C.num_sample, 1, 1) - gen_syn, 2))
                            gen_optimizer.zero_grad()
                            gen_loss.backward()
                            gen_optimizer.step() 
                        else:
                            nt = status.shape[0]
                            for i in range(self.C.num_sample * self.C.num_gen_iter):

                                gen_loss = torch.mean(torch.pow(des_syn[nt*i : nt*(i+1), self.C.his_size:, :2].detach() - self.gen_model.sample(status[..., self.C.his_size:, :], 
                                extend[..., self.C.his_size:, :])[..., :2], 2))
                                gen_optimizer.zero_grad()
                                gen_loss.backward()
                                gen_optimizer.step()
                        gen_scheduler.step()
                        self.tb_writer.add_scalar("general/lr_gen", gen_scheduler.get_last_lr()[0], epoch)
                        res_gen = util_torch.EvaluationMetric("train/gen")
                        res_gen.calculate(status[...,self.C.his_size:,:], gen_syn[...,self.C.his_size:,:])
                        res_gen.print_save(self.tb_writer, epoch)
                        self.tb_writer.add_scalar("general/g_loss", float(gen_loss), epoch)
                else:
                    gen_loss = -1

                res_des = util_torch.EvaluationMetric("train/des")
                res_des.calculate(status[...,self.C.his_size:,:], des_syn[...,self.C.his_size:,:])
                res_des.print_save(self.tb_writer, epoch)
                self.pbar.update(epoch % self.C.eval_interval)
                self.tb_writer.add_scalar("general/d_loss", float(des_loss), epoch)
                self.tb_writer.add_scalar("general/des_obs_cost", float(torch.mean(obs_cost)), epoch)
                self.tb_writer.add_scalar("general/des_syn_cost", float(torch.mean(syn_cost)), epoch)


                if epoch % self.C.eval_interval == 0:

                    res_gen = util_torch.EvaluationMetric("val/gen")
                    res_des = util_torch.EvaluationMetric("val/des")
                    for i_batch, b_data in enumerate(validate_loader):

                        status = b_data['status'].type(torch.float32)
                        extend = b_data['extend'].type(torch.float32)
                        gen_syn, des_syn = self.cooperative_process_multi(status, extend)
                        
                        res_gen.calculate(status[...,self.C.his_size:,:], gen_syn[...,self.C.his_size:,:])
                        res_des.calculate(status[...,self.C.his_size:,:], des_syn[...,self.C.his_size:,:])
                        if i_batch == 0:
                            fig, fig_c = util_torch.visualize_tray_batch_fig(status[...,self.C.his_size:,:].cpu().data.numpy(), 
                                des_syn[...,self.C.his_size:,:].cpu().data.numpy(), extend=extend.cpu().data.numpy())
                            self.tb_writer.add_figure("val/trajectory", fig, epoch)
                            if fig_c is not None: 
                                self.tb_writer.add_figure("val/control", fig_c, epoch)
                            break

                    res_train_gen = res_gen.print_save(self.tb_writer, epoch) if self.gen_model is not None else (0,0,0) 
                    res_train_des = res_des.print_save(self.tb_writer, epoch)
                    res_gen = util_torch.EvaluationMetric("test/gen")
                    res_des = util_torch.EvaluationMetric("test/des")
                    res_traj = []
                    for i_batch, b_data in enumerate(test_loader):

                        status = b_data['status'].type(torch.float32)
                        extend = b_data['extend'].type(torch.float32)
                        gen_syn, des_syn = self.cooperative_process_multi(status, extend, -1 if i_batch>0 else epoch)
                        res_gen.calculate(status[...,self.C.his_size:,:], gen_syn[...,self.C.his_size:,:])
                        res_des.calculate(status[...,self.C.his_size:,:], des_syn[...,self.C.his_size:,:])
                        res_traj.append(des_syn.cpu().data.numpy())
                        if i_batch == 0:
                            fig, fig_c = util_torch.visualize_tray_batch_fig(status[...,self.C.his_size:,:].cpu().data.numpy(), 
                                des_syn[...,self.C.his_size:,:].cpu().data.numpy(), extend=extend.cpu().data.numpy())
                            self.tb_writer.add_figure("test/trajectory", fig, epoch)
                            if fig_c is not None: 
                                self.tb_writer.add_figure("test/control", fig_c, epoch)
                                          
                    res_test_gen = res_gen.print_save(self.tb_writer, epoch) if self.gen_model is not None else (0,0,0) 
                    res_test_des = res_des.print_save(self.tb_writer, epoch)
                    self.tb_writer.flush()
                    if self.gen_model is not None:
                        print("Epoch #%d: \t Fast thinking: RMSE = %.3f / %.3f; ADE = %.3f / %.3f; FDE = %.3f / %.3f; g_loss: %.5f" % (epoch, res_train_gen[0], res_test_gen[0], res_train_gen[1], res_test_gen[1], res_train_gen[2], res_test_gen[2], float(gen_loss)))
                    print("\t\t Slow thinking: RMSE = %.3f / %.3f; ADE = %.3f / %.3f; FDE = %.3f / %.3f; d_loss: %.5f"% (res_train_des[0], res_test_des[0], res_train_des[1], res_test_des[1], res_train_des[2], res_test_des[2], float(des_loss)))
                    np.save("output/traj/%s.npy" % (self.C.category), np.concatenate(res_traj))
                    to_save = [self.des_model.state_dict()]
                    if self.gen_model is not None:
                        to_save.append(self.gen_model.state_dict())
                    torch.save(to_save, "output/model/%s.pt" % self.C.category)
        
                epoch += 1
                if epoch > self.C.num_epochs:
                    break 
            if epoch > self.C.num_epochs:
                    break 
        self.pbar.finish()
        util_torch.PushNotification().post_text("[CIOC] %s finished!" % self.C.category)
        
    def test_delta(self, pt_path, test_data):

        N_test = len(test_data)
        self.des_model, self.gen_model = torch.load(pt_path)
        # self.gen_model = None
        # self.C.output_dir += "_evaluate"
        # preload_theta = Tensor([1.00000000e-02, 1.58858228e-04, 6.06721685e-02, 2.63659442e-05,
        # 1.46766149e+03, 1.19869357e+03, 4.26824864e-03, 4.81507192e+01,
        # 1.11108007e+02, 9.93697501e+01])
        # self.des_model.weight = (self.des_model.weight / list(self.des_model.parameters())[0] * 100).detach()[0] * 100
        # print(self.des_model.weight)
        # self.des_model.weight[0:4] = 0
        self.des_model.weight[4:6] = 0
        # self.des_model.weight[7] = 10
        print(self.des_model.weight)
        print(list(self.des_model.parameters())[0])

        # basic_check(test_data[:]['status'], self.car_dynamic.car_dynamic_pt, self.C)

        delta_list = [0, 0.05, 0.08, 0.1, 0.12, 0.15, 0.2, 0.4, 0.8]
        x = np.array(range(self.C.sample_steps)) + 1
        rmse_list = []
        for i, delta in enumerate(delta_list):
            self.C.delta = delta

            log_saver = util_torch.LogSaver(self.C.sample_steps, self.C.future_size, self.C.single, (self.C.des_net=='linear'))
            test_loader = tdata.DataLoader(test_data, batch_size=1000, shuffle=False)
            for i_batch, b_data in enumerate(test_loader):

                status = b_data['status'].type(torch.float32)
                extend = b_data['extend'].type(torch.float32)
                t_init, x_full = self.cooperative_process(status, extend, test_epoch=i)
                for t_step in range(self.C.sample_steps):
                    status_syn = x_full[t_step]
                    error = util_torch.l2_error(status[...,self.C.his_size:,:], status_syn[...,self.C.his_size:,:]).cpu().data.numpy()
                    # gen_error = util_torch.l2_error(status[...,self.C.his_size:,:], t_init[...,self.C.his_size:,:]).cpu().data.numpy()[:, -1]
                    log_saver.save(t_step, 0, error, train=False)
                # if i_batch == 0:
                #     util_torch.visualize_tray_batch(status[...,self.C.his_size:,:].cpu().data.numpy(), status_syn[...,self.C.his_size:,:].cpu().data.numpy(),
                #                         self.C.output_dir + '/trajectory_%.3f' % (delta), extend=extend.cpu().data.numpy())
            print("Delta = %f" % delta)
            # print(log_saver.test_rmse)
            rmse_list.append(log_saver.test_rmse)
            print(log_saver.mesg(self.C.sample_steps-1))

            plt.clf()
            np.save("output/eval/%s_ress.npy" % self.C.category, np.stack(rmse_list))
            for i in range(len(rmse_list)):
                plt.plot(x[0:-1:2], np.array(rmse_list[i])[0:-1:2], label="delta = %.2f" % (delta_list[i]))
            plt.legend(loc=1, prop={'size': font_size})
            plt.xlabel("Number of steps", fontsize=font_size)
            plt.ylabel("RMSE", fontsize=font_size)
            plt.tight_layout()
            plt.grid(color='grey', linestyle='-', linewidth=0.5)
            plt.ylim([np.min(np.array(rmse_list)) - 0.05, rmse_list[i][0] + 0.1])
            plt.savefig("output/eval/%s_rmse_stepsize.png" % self.C.category)
            plt.savefig("output/eval/%s_rmse_stepsize.eps" % self.C.category)
            plt.clf()

    def test_multi(self, pt_path, test_data):

        N_test = len(test_data)
        self.des_model, _ = torch.load(pt_path)
        self.C.output_dir += "_evaluate"
        # preload_theta = Tensor([1.00000000e-02, 1.58858228e-04, 6.06721685e-02, 2.63659442e-05,
        # 1.46766149e+03, 1.19869357e+03, 4.26824864e-03, 4.81507192e+01,
        # 1.11108007e+02, 9.93697501e+01])
        # self.des_model.weight = (self.des_model.weight / list(self.des_model.parameters())[0] * 100).detach()[0] * 100
        # print(self.des_model.weight)
        # self.des_model.weight[0:4] = 0
        self.des_model.weight[4:6] = 0
        self.des_model.weight[7] = 10
        print(self.des_model.weight)
        print(list(self.des_model.parameters())[0])

        util_torch.basic_check(test_data[:]['status'], self.car_dynamic.car_dynamic_pt, self.C)

        log_saver = util_torch.LogSaver(self.C.sample_steps, self.C.future_size, self.C.single, (self.C.des_net=='linear'))
        test_loader = tdata.DataLoader(test_data, batch_size=5000, shuffle=False)
        for i_batch, b_data in enumerate(test_loader):

            status = b_data['status'].type(torch.float32)
            extend = b_data['extend'].type(torch.float32)
            t_init, x_full = self.cooperative_process(status, extend, test_epoch=i)
            for t_step in range(self.C.sample_steps):
                status_syn = x_full[t_step]
                error = util_torch.l2_error(status[...,self.C.his_size:,:], status_syn[...,self.C.his_size:,:]).cpu().data.numpy()
                # gen_error = util_torch.l2_error(status[...,self.C.his_size:,:], t_init[...,self.C.his_size:,:]).cpu().data.numpy()[:, -1]
                log_saver.save(t_step, 0, error, train=False)
            if i_batch == 0:
                util_torch.visualize_tray_batch(status[...,self.C.his_size:,:].cpu().data.numpy(), status_syn[...,self.C.his_size:,:].cpu().data.numpy(),
                                    self.C.output_dir + '/trajectory_%.3f' % (delta), extend=extend.cpu().data.numpy())
        print("Delta = %f" % 2**delta)
        # print(log_saver.test_rmse)
        rmse_list.append(log_saver.test_rmse)
        print(log_saver.mesg(self.C.sample_steps-1))

def main_test(opt):

    # with open("output/%s/config.pkl" % opt.category, "rb") as conf:
    #     opt = pickle.load(conf)
    opt.print_level = 100
    pt_path = os.path.join('./output/model/%s.pt' % opt.category)
    np.random.seed(opt.seed)
    np.random.seed(opt.seed)
    train_data, test_data = load_data(opt)
    model = IOC_Cooperative(opt)
    # model.des_model.weight = model.normalize_weight(train_data)
    model.test_delta(pt_path, test_data)
