import numpy as np
import torch
try:
    import tianshou as ts 
    from tianshou.utils.net.continuous import ActorProb, Critic
    from tianshou.utils.net.common import Net
except:
    raise ImportError

# Section 1 : Cost function network series with input status and extend (des_net)

# 1.1 : Linear Model
class ModelLinear(torch.nn.Module):

# Basic linear model

    def __init__(self, config, cost_function):
        super(ModelLinear, self).__init__()
        self.out = torch.nn.Linear(config.feature_size, 1)
        self.out.weight = torch.nn.Parameter(torch.randn((1, config.feature_size)) * config.std_para)
        self.cost_func = cost_function
        self.weight = torch.ones(config.feature_size, device=config.device)

    def forward(self, status, extend):
        return torch.sum(self.out(self.cost_func(status, extend) * self.weight), 1)

# 1.2.1 : MLP model --- 2 layer
class ModelMLP2(torch.nn.Module):

    def __init__(self, config, cost_function):
        super(ModelMLP2, self).__init__()
        
        self.out = torch.nn.Sequential(
                        torch.nn.Linear(config.feature_size, config.hidden_size),
                        torch.nn.LeakyReLU(inplace=True),
                        torch.nn.Linear(config.hidden_size, 1))
        self.cost_func = cost_function
        self.weight = torch.ones(config.feature_size, device=config.device)
    def forward(self, status, extend):
        return torch.sum(self.out(self.cost_func(status, extend) * self.weight), 1)

# 1.2.2 : MLP model --- 3 layer
class ModelMLP3(torch.nn.Module):

    def __init__(self, config, cost_function):
        super(ModelMLP3, self).__init__()
        self.out = torch.nn.Sequential(
                        torch.nn.Linear(config.feature_size, config.hidden_size),
                        torch.nn.LeakyReLU(inplace=True),
                        torch.nn.Linear(config.hidden_size, config.hidden_size // 2),
                        torch.nn.LeakyReLU(inplace=True),
                        torch.nn.Linear(config.hidden_size // 2, 1))
        self.cost_func = cost_function
        self.weight = torch.ones(config.feature_size, device=config.device)

    def forward(self, status, extend):
        return torch.sum(self.out(self.cost_func(status, extend) * self.weight), 1)

# 1.2.3 : MLP model --- 4 layer
class ModelMLP4(torch.nn.Module):

    def __init__(self, config, cost_function):
        super(ModelMLP4, self).__init__()

        self.out = torch.nn.Sequential(
                        torch.nn.Linear(config.feature_size, config.hidden_size),
                        torch.nn.LeakyReLU(inplace=True),
                        torch.nn.Linear(config.hidden_size, config.hidden_size),
                        torch.nn.LeakyReLU(inplace=True),
                        torch.nn.Linear(config.hidden_size, config.hidden_size),
                        torch.nn.LeakyReLU(inplace=True),
                        torch.nn.Linear(config.hidden_size, 1))
        self.cost_func = cost_function
        self.weight = torch.ones(config.feature_size, device=config.device)

    def forward(self, status, extend):
        return torch.sum(self.out(self.cost_func(status, extend) * self.weight), 1)

# 1.2.4 : MLP model --- 5 layer
class ModelMLP5(torch.nn.Module):

    def __init__(self, config, cost_function):
        super(ModelMLP5, self).__init__()

        self.out = torch.nn.Sequential(
                        torch.nn.Linear(config.feature_size, config.hidden_size),
                        torch.nn.LeakyReLU(inplace=True),
                        torch.nn.Linear(config.hidden_size, config.hidden_size),
                        torch.nn.LeakyReLU(inplace=True),
                        torch.nn.Linear(config.hidden_size, config.hidden_size),
                        torch.nn.LeakyReLU(inplace=True),
                        torch.nn.Linear(config.hidden_size, config.hidden_size),
                        torch.nn.LeakyReLU(inplace=True),
                        torch.nn.Linear(config.hidden_size, 1))
        self.cost_func = cost_function
        self.weight = torch.ones(config.feature_size, device=config.device)

    def forward(self, status, extend):
        return torch.sum(self.out(self.cost_func(status, extend) * self.weight), 1)

# 1.3 : CNN model
class ModelCNN40(torch.nn.Module):

    def __init__(self, config, cost_function):
        super(ModelCNN40, self).__init__()
        self.out = torch.nn.Sequential(
                        torch.nn.Conv1d(config.feature_size, 16, 4, 2),
                        torch.nn.ReLU(inplace=True),
                        torch.nn.Conv1d(16, 32, 4, 2, 1),
                        torch.nn.ReLU(inplace=True),
                        torch.nn.Conv1d(32, 64, 4, 2, 1),
                        torch.nn.ReLU(inplace=True),
                        torch.nn.Conv1d(64, 128, 4, 1),
                        torch.nn.ReLU(inplace=True))
        self.mlp = torch.nn.Linear(128, 1)
        self.cost_func = cost_function
        self.weight = torch.ones(config.feature_size, device=config.device)

    def forward(self, status, extend):
        sub_cost = self.cost_func(status, extend) * self.weight
        res = self.out(sub_cost.permute(0, 2, 1))
        return self.mlp(res.squeeze())

class ModelCNN40_big(torch.nn.Module):

    def __init__(self, config, cost_function):
        super(ModelCNN40_big, self).__init__()
        self.out = torch.nn.Sequential(
                        torch.nn.Conv1d(config.feature_size, 32, 4, 2),
                        torch.nn.LeakyReLU(inplace=True),
                        torch.nn.Conv1d(32, 64, 4, 2, 1),
                        torch.nn.LeakyReLU(inplace=True),
                        torch.nn.Conv1d(64, 128, 4, 2, 1),
                        torch.nn.LeakyReLU(inplace=True),
                        torch.nn.Conv1d(128, 256, 4, 1),
                        torch.nn.LeakyReLU(inplace=True))
        self.mlp = torch.nn.Linear(256, 1)
        self.cost_func = cost_function
        self.weight = torch.ones(config.feature_size, device=config.device)

    def forward(self, status, extend):
        sub_cost = self.cost_func(status, extend) * self.weight
        res = self.out(sub_cost.permute(0, 2, 1))
        return self.mlp(res.squeeze())

# 1.4 Network used in GAIL

class ModelGAIL(torch.nn.Module):

    def __init__(self, config, cost_function):

        super(ModelGAIL, self).__init__()
        self.hidden_size = config.hidden_size
        self.base_cost = cost_function
        activation_fn = torch.nn.ReLU()
        self.model = torch.nn.Sequential(
            torch.nn.Linear(config.feature_size, config.hidden_size),
            activation_fn, 
            torch.nn.Linear(config.hidden_size, config.hidden_size),
            activation_fn, 
            torch.nn.Linear(config.hidden_size, 1),
            torch.nn.Sigmoid())
        self.loss_fct = torch.nn.BCEWithLogitsLoss()

    def forward(self, state, extend, labels=None):

        # Input size : batch * state_size 
        output = self.model(self.base_cost(state, extend)).squeeze()
        if labels is not None: 
            loss = self.loss_fct(output, labels)
            return loss, output
        return output

# Section 2 : generator network series 

# 2.1 : fast-thinking / ABP model 

class GeneratorTime(torch.nn.Module):

    def __init__(self, config, dynamic, activation=torch.nn.ReLU, cost_function=lambda x:x):
        super(GeneratorTime, self).__init__()
        self.out = torch.nn.Sequential(
                        torch.nn.Linear(config.state_size + config.extend_size + config.latent_size, 64),
                        activation(),
                        torch.nn.Linear(64, 16),
                        activation(),
                        torch.nn.Linear(16, 8),
                        activation(),
                        torch.nn.Linear(8, config.action_size),
                        torch.nn.Tanh())
        self.latent_size = config.latent_size
        self.cost_func = cost_function
        self.state_size = config.state_size
        self.output_size = config.state_size + config.action_size
        self.dynamic = dynamic

    # Status : batch_size * traj_size * status_size; extend : batch_size * traj_size * status_size
    def forward(self, z, status, extend):
        traj_size = status.shape[1]
        res = torch.zeros((status.shape[0], traj_size, self.output_size))
        hx = status[..., 0, :self.state_size]
        for t in range(traj_size):
            ta = self.out(torch.cat((hx, z[..., t, :], extend[..., t, :]), 1))
            res[..., t, :self.state_size] = hx
            res[..., t, self.state_size:] = ta
            hx = self.dynamic(hx, ta)
        return res

    def sample(self, status, extend):
        z = torch.randn((status.shape[0], status.shape[1], self.latent_size), device=status.device)
        return self.forward(z, status, extend)

class GeneratorOri(torch.nn.Module):

    def __init__(self, config, dynamic, activation=torch.nn.Tanh):
        super(GeneratorOri, self).__init__()
        self.out = torch.nn.Sequential(
                        torch.nn.Linear(config.state_size + config.extend_size + config.latent_size, 32),
                        activation(),
                        torch.nn.Linear(32, 16),
                        activation(),
                        torch.nn.Linear(16, config.action_size),
                        torch.nn.Tanh())
        self.latent_size = config.latent_size
        self.output_size = config.state_size + config.action_size
        self.dynamic = dynamic

    # Status : batch_size * traj_size * status_size; extend : batch_size * traj_size * status_size
    def forward(self, z, status, extend):
        traj_size = status.shape[1]
        res = torch.zeros((status.shape[0], traj_size, self.output_size))
        hx = status[..., 0, :6]
        for t in range(traj_size):
            ta = self.out(torch.cat((hx, z[..., t, :], extend[..., t, :]), 1))
            res[..., t, :6] = hx
            res[..., t, 6:] = ta
            hx = self.dynamic(hx, ta)
        return res

    def sample(self, status, extend):
        z = torch.randn((status.shape[0], status.shape[1], self.latent_size), device=status.device)
        return self.forward(z, status, extend)
